import pytest # noqa: F401
import numpy as np
import pandas as pd
import os
print(os.getcwd())

from tgcnn_act_graph import utils, max_act_diff, figures, calculations  # noqa: E402

test_pats = np.array([[[[0, 0, 0], 
                          [0, 0, 3], 
                          [0, 0, 0]], [[0, 0, 0],
                                       [0, 0, 0],
                                       [0, 4, 4]], [[0, 0, 0], 
                                                    [0, 2, 0], 
                                                    [0, 2, 0]], [[0, 0, 0], 
                                                                 [5, 0, 5], 
                                                                 [0, 0, 0]]],                        
                        [[[1, 0, 0], 
                          [1, 0, 0], 
                          [1, 0, 0]], [[0, 0, 0],
                                       [0, 8, 0],
                                       [0, 0, 0]], [[0, 11, 0], 
                                                    [0, 0, 0], 
                                                    [0, 11, 0]], [[0, 0, 30], 
                                                                 [0, 0, 30], 
                                                                 [0, 0, 30]]],                       
                          
                          [[[0, 0, 9], 
                          [0, 0, 0], 
                          [0, 0, 0]], [[0, 0, 0],
                                       [7, 0, 0],
                                       [7, 0, 0]], [[0, 0, 0], 
                                                    [0, 0, 2], 
                                                    [0, 0, 0]], [[0, 0, 0], 
                                                                 [1, 0, 0], 
                                                                 [0, 0, 0]]],                          
                          [[[1, 1, 1], 
                          [0, 0, 0], 
                          [0, 0, 0]], [[0, 0, 1],
                                       [0, 0, 1],
                                       [0, 0, 0]], [[0, 0, 0], 
                                                    [0, 1, 0], 
                                                    [0, 1, 0]], [[0, 2, 0], 
                                                                 [0, 0, 0], 
                                                                 [0, 2, 0]]]]
                        
                        )

test_filts = np.array([[[[0, 0, 1], 
                    [0, 1, 0], 
                    [0, 0, 1]], [[1, 1, 0], 
                                 [0, 0, 0], 
                                 [0, 0, 0]]], 
                   [[[0, 1, 0], 
                    [0, 1, 0], 
                    [0, 0, 1]], [[1, 0, 0], 
                                 [0, 0, 0], 
                                 [0, 0, 0]]]])


test_pat = np.array([[[0, 0, 0], 
                        [0, 0, 3], 
                        [0, 0, 0]], [[0, 0, 0],
                                    [0, 0, 0],
                                    [0, 4, 4]], [[0, 0, 0], 
                                                [0, 2, 0], 
                                                [0, 2, 0]], [[0, 0, 0], 
                                                                [5, 0, 5], 
                                                                [0, 0, 0]]])

test_filt = np.array([[[0, 0, 1], 
                        [0, 1, 0], 
                        [0, 0, 1]], [[1, 1, 0], 
                                        [0, 0, 0], 
                                        [0, 0, 0]]])

labels = [0, 1, 1, 0] #hip_replacement labels
def test_y_coord_list_gen():
    """Test the lists generated for the y coordinates of the nodes
    are generating as expected.
    """
    auto_seq = utils.generate_pos_sequence(10)

    test_ans1 = auto_seq[0]
    true_ans1 = [0]

    assert test_ans1 == true_ans1, "The two lists should be equal"

    test_ans2 = auto_seq[4]
    true_ans2 = [-2, -1, 0, 1, 2]

    assert test_ans2 == true_ans2, "The two lists should be equal"

    test_ans3 = auto_seq[7]
    true_ans3 = [-4, -3, -2, -1, 1, 2, 3, 4]

    assert test_ans3 == true_ans3, "The two lists should be equal"


def test_pos_dict():
    """ Test that the y coordinate dictionaries have the same keys and values.
    The dictionaries may not match exactly as the y coordinate is dependent on the order of the
    DataFrame.
    """
    # How pos dictionary should look:
    pos_true = {
        '1_v0': (0, 0),
        '2_v1': (1, 0),
        '1_v2': (2, 1),
        '2_v2': (2, -1),
        '1_v3': (3, 0),
        '0_v4': (4, 1),
        '2_v4': (4, -1)
    }

    # Code chunk to get y position dictionary
    mean_activation_df = max_act_diff.max_act_diff_calc('Hip Replacement', test_pats, test_filts, labels, verbose=False, show_plot=False)
    max_act_filt = utils.get_max_act_filt(mean_activation_df, test_filts)
    act_graph = calculations.get_act_graph_array(test_pat, max_act_filt)
    edges_df = utils.create_edges_df(test_pat, act_graph)
    pos_df = utils.create_position_df(edges_df)
    pos_list = utils.generate_pos_sequence(pos_df['max_codes_per_visit'].max())
    pos_df = utils.map_y_coord_to_node(pos_df, pos_list)
    pos_test = utils.create_pos_dict(pos_df)


    pos_true_keys = set(pos_true.keys())
    pos_test_keys = set(pos_test.keys())

    pos_true_values = set(pos_true.values())
    pos_test_values = set(pos_test.values())

    assert pos_true_keys == pos_test_keys, "The keys of dictionaries of pos_true and pos_test should be equal"
    assert pos_true_values == pos_test_values, "The values of dictionaries of pos_true and pos_test should be equal"


def test_edge_weight():
    """Test that the weight of the activated edge in the test patient is calculated correctly.
    """

    true_weight1 = 2.0
    
    mean_activation_df = max_act_diff.max_act_diff_calc('Hip Replacement', test_pats, test_filts, labels, verbose=False, show_plot=False)
    max_act_filt = utils.get_max_act_filt(mean_activation_df, test_filts)
    act_graph = calculations.get_act_graph_array(test_pat, max_act_filt)
    edges_df = utils.create_edges_df(test_pat, act_graph)

    test_weight1 = edges_df.loc[edges_df['activated']==1, 'weight'].astype(float).values

    assert true_weight1 == test_weight1, "The weight of this edge should be 2."

def test_extract_visit_num():
    """Test the extracted visit number is the visit number and not the node number.
    """
    string = "5_v2"
    true_visit_num = 2
    false_visit_num = 5
    test_visit_num = utils.extract_visit_number(string)

    assert true_visit_num == test_visit_num, "True visit number should equal test visit number"
    assert false_visit_num != test_visit_num, "True visit number should equal test visit number not EHR code number"

# TODO: Test that the repeat_array_fractional function is repeating correctly


# def test_repeat_array_fractional():
#     """Test to make sure filter is repeating correctly.
#     """
    


import numpy as np
import pandas as pd


def select_patient(input_tensors:np.array, num:int):
    """Select a single patient. Choose an integer as the patient number.

    Args:
        input_tensors (np.array): 4D numpy array representing patients EHR graphs.
        num (int): patient number to choose.

    Returns:
        np.array: 3D array representing an individuals EHR graph.
    """
    return input_tensors[num]

def extract_visit_number(s:str):
    """Extract the visit number from a string

    Args:
        s (str): string to split

    Returns:
        int: visit number
    """
    return int(s.split('_')[1][1:])

def get_max_act_filt(mean_activation_df:pd.DataFrame, filters:np.array):
    """Using the dataframe of acitvation difference select the filter with the largest difference.

    Args:
        mean_activation_df (pd.DataFrame): dataframe with difference in activation between positive and negative class 
        for each filter.
        filters (np.array): 4D array with 3D filters.

    Returns:
        np.array: 3D array representing one filter.
    """

    max_idx = mean_activation_df['Difference'].idxmax()
    max_act_filt_num = mean_activation_df.loc[max_idx, 'Filter']
    max_act_filt =  filters[max_act_filt_num-1] # minus 1 as we don't have a filter called 0
    print(f"Filter {max_act_filt_num} had the largest difference in activation.")
    return max_act_filt



def create_edges_df(patient_graph:np.array, act_graph:np.array):
    """Create a DataFrame of the edges of the patient graph, including the start and end nodes
    whether the edge is activated (more than 0), the 'weight' of the activation, and the time 
    between visits.

    Args:
        patient_graph (np.array): 3D numpy array showing the patients health codes over time.
        act_graph (np.array): 3D numpy array showing graph activation.

    Returns:
        DataFrame: with columns start_node, end_node, activated, weight, time_between.
    """
    num_edges = np.count_nonzero(patient_graph)
    edges_df = pd.DataFrame(columns=['start_node', 'end_node', 'activated', 'weight', 'time_between'], index=range(num_edges))
    num_nodes = patient_graph.shape[1]
    timesteps = patient_graph.shape[0]

    row_num = 0
    for t in range(timesteps):
        for i in range(num_nodes):
            for j in range(num_nodes):
                if patient_graph[t, i, j] != 0:
                    if t == 0:
                        start_node_v = 0
                        end_node_v = 1
                    else:
                        start_node_v = t
                        end_node_v = t+1
                    
                    edges_df.at[row_num, 'end_node'] = f'{j}_v{end_node_v}' #[row num, col name]
                    edges_df.at[row_num, 'weight'] = act_graph[t, i, j]
                    edges_df.at[row_num, 'time_between'] = patient_graph[t, i, j] #add a column for the edge label too (time between from original graph)
                    edges_df.at[row_num, 'start_node'] = f'{i}_v{start_node_v}' #[row num, col name]
                    
                    row_num += 1


    # activated_graph col is more than 0, the activated column = 1, otherwise it is 0
    edges_df['activated'] = edges_df['weight'].apply(lambda x: 0 if x == 0 else 1)  
    # change 0 weights to 0.5 so we can still see them on the graph figure
    edges_df['weight'] = edges_df['weight'].apply(lambda x: 0.5 if x == 0 else x)
    return edges_df


def create_position_df(edges_df:np.array):
    """Create a DataFrame to find the number of nodes per visit.

    Args:
        edges_df (pd.DataFrame): DataFrame with columns start_node, end_node, activated, weight, time_between.

    Returns:
        pd.DataFrame: with columns showing node name (node), visit number (x), node number in visits (cumulative count),
        and maximum codes per visit.
    """
    pos_df = edges_df[['start_node', 'end_node']].stack().drop_duplicates().reset_index(drop=True)
    pos_df = pos_df.to_frame(name='node')
    pos_df['x'] = pos_df['node'].apply(extract_visit_number)
    pos_df['cumulative_count'] = pos_df.groupby('x').cumcount()
    pos_df['max_codes_per_visit'] = pos_df.groupby('x')['cumulative_count'].transform('max') + 1

    return pos_df


def generate_pos_sequence(x:int):
    """Generate a list of lists to get y coordinate positions for the nodes
     based on the number of events recorded per visit.

    Args:
        x (int): maximum number of nodes in any one visit

    Returns:
        List: List of lists of y coordinates index mapping to the max nodes for
               each visit.
    """
    sequence = []
    for i in range(x):
        if i % 2 == 0:  # Even index, include zero
            sequence.append(list(range(-i // 2, i // 2 + 1)))
        else:  # Odd index, exclude zero
            sublist = list(range(-(i // 2 + 1), i // 2 + 2))
            sublist.remove(0)
            sequence.append(sublist)
    return sequence


def get_pos_y_value_per_node(row, pos_list:list):
    """Get the y position for each node. Use max_codes_per_visit column to select the sublist 
    and the cumulative_count to get the position from the sublist.

    Args:
        row (pd.Series): row of the DataFrame

    Returns:
        int: y coordinate position 
    """
    cum_count = row['cumulative_count']
    max_codes = row['max_codes_per_visit']
    return pos_list[max_codes - 1][cum_count]

def map_y_coord_to_node(pos_df:pd.DataFrame, pos_list:list):
    """Map the y coordinates to the relevant node and correct row.

    Args:
        pos_df (pd.DataFrame): columns with node name and x coordinate.
        pos_list (list of lists): List of lists of y coordinates index mapping to the max nodes for
               each visit.

    Returns:
        pd.DataFrame: dataframe with x and y coordinates for node plotting.
    """
    pos_df['y'] = pos_df.apply(lambda row: get_pos_y_value_per_node(row, pos_list), axis=1)
    return pos_df

def create_pos_dict(pos_df:pd.DataFrame):
    """Make a dictionary with the node name as the key and the x and y coordinates 
    as a tuple value.

    Args:
        pos_df (pd.DataFrame): dataframe with columns for the node name, x coordinates,
        and y coordinates.

    Returns:
        dict: dictionary of node: (x,y)
    """
    # the visit number is x and the y value is the number of nodes with the same visit number
    pos = pos_df.set_index('node')[['x', 'y']].apply(tuple, axis=1).to_dict()
    return pos


import pandas as pd
import numpy as np
from tgcnn_act_graph import calculations, figures

def create_max_act_df(class_name:str, pat_graphs:np.array, filters:np.array, labels:list, verbose:bool):
    """Calculate the maximum activation from each filter on each patient graph. Running a filter over
    each patient graph and getting the max. 
    Assumes a stride length of one.

    Args:
        class_name (str): name to describe prediction outcome.
        pat_graphs (np.array): 4D array containing x 3D patient graphs.
        filters (np.array): 4D array containing x filters.
        labels (list): list of binary values representing positive or negative outcomes.
        verbose (bool): print or not to print extra dataframes or print statements.

    Raises:
        ValueError: if numpy array isn't 4D.

    Returns:
        pd.DataFrame: dataframe with columns for filter number, maximum activation and chosen class_name string.
    """
    if pat_graphs.ndim != 4:
        raise ValueError("The input array must be 4-dimensional.") # 4D to have multiple patients 3D graph representations
    if filters.ndim != 4:
        raise ValueError("The filters array must be 4-dimensional.") # 4D to have multiple 3D filters

    columns = [class_name, 'Max Activation', 'Filter']
    time_steps = pat_graphs.shape[1]
    num_pats = pat_graphs.shape[0]
    num_filters = filters.shape[0]
    
    # Construct empty DataFrame
    num_rows = num_pats * num_filters
    df_index = 0
    df = pd.DataFrame(columns=columns, index=range(num_rows))

    for filter_num in range(0, num_filters):
        f = filters[filter_num]
        
        # first run every patient for one filter
        for patient_num in range(0, num_pats):
            df.iloc[df_index, 0] = labels[patient_num]
            input_tensor = pat_graphs[patient_num]
            # Run a single filter over the graph
            max_activation_value = 0
            # this might need to change when the stride isn't 1 - TODO add this as an action to github 
            for i in range(0, time_steps-1): 
                first_slice = input_tensor[i:f.shape[0]+i, :, :]
                filter_times_slice = first_slice * f
                window_sum = np.sum(filter_times_slice)
                activation = calculations.leaky_relu(window_sum, 0.01) 
                if activation > max_activation_value:
                    max_activation_value = activation
            df.iloc[df_index, 1] = max_activation_value
            df.iloc[df_index, 2] = filter_num+1
            df_index += 1

            if verbose:
                print(f"The maximum activation for patient {patient_num} and filter {filter_num} = {max_activation_value}")
    
    return df

def max_act_diff_calc(class_name:str, pat_graphs:np.array, filters:np.array, labels:list, verbose:bool, show_plot:bool):
    """Calculate the max difference between the classes for each filter and return in a dataframe.

    Args:
        class_name (str): name to describe prediction outcome.
        pat_graphs (np.array): 4D array containing x 3D patient graphs.
        filters (np.array): 4D array containing x filters.
        labels (list): list of binary values representing positive or negative outcomes.
        verbose (bool): print or not to print extra dataframes or print statements.
        show_plot (bool): display graph of max filter activations. 

    Returns:
        pd.DataFrame: difference between graph activation in both classes.
    """

    max_act_per_filt_df = create_max_act_df(class_name, pat_graphs, filters, labels, verbose)
    mean_activation = max_act_per_filt_df.groupby(['Filter', class_name])['Max Activation'].mean()
    mean_activation_df = mean_activation.to_frame()
    mean_activation_df.reset_index(inplace=True)

    # Calculate the difference between the two class type rows for each Filter
    mean_activation_df['Difference'] = mean_activation_df.groupby('Filter')['Max Activation'].diff()
    mean_activation_df['Difference'] = mean_activation_df['Difference'].abs()

    mean_activation_df = mean_activation_df[['Filter', 'Difference']].dropna().reset_index(drop=True)

    mean_activation_df['Filter'] = mean_activation_df['Filter'].astype(int)
    mean_activation_df['Difference'] = mean_activation_df['Difference'].astype(float)

    if verbose:
        print(mean_activation_df)
    if show_plot:
        figures.plot_activation_difference(mean_activation_df['Filter'], mean_activation_df['Difference'])

    return mean_activation_df




import numpy as np


def repeat_array_fractional(array: np.ndarray, repeats: float):
    """
    Repeats the filter array a specified non-integer number of times.
    This is because the graph may divide fractionally compared to the filter size.
    
    Parameters:
    array (np.ndarray): The array to repeat.
    repeats (float): The number of times to repeat the array.
    
    Returns:
    np.ndarray: The repeated array.
    """
    # Separate the integer and fractional parts
    int_part = int(repeats)
    frac_part = repeats - int_part

    # Repeat the array for the integer part
    repeated_array_int = np.tile(array, (int_part, 1, 1))

    # Calculate the number of rows to take for the fractional part
    num_rows = array.shape[0]
    frac_rows = int(num_rows * frac_part)

    # Take the fractional part from the array
    if frac_rows > 0:
        repeated_array_frac = array[:frac_rows, :, :]
        # Concatenate the integer part and the fractional part
        result_array = np.concatenate((repeated_array_int, repeated_array_frac), axis=0)
    else:
        result_array = repeated_array_int

    return result_array

def leaky_relu(x:float, alpha:int):
    """If the activation is negative multiply it by alpha,
    otherwise retain the alpha value.
    This is to avoid 0 gradients.

    Args:
        x (float): activation sum
        alpha (int): hyperparameter to select or tune

    Returns:
        float: activation sum with leaky ReLU applied
    """
    return max(alpha * x, x) # if it's negative multiply by 0.01


def get_act_graph_array(pat_graph: np.array, max_act_filt: np.array):
    """Get the activated_graph which is the graph element-wise multiplied by the 'sliding window'. 

    The stride must be the same as the filter.shape[0] so that slices don't overlap (otherwise we
    get multiple mappings).

    The filter is repeated to be the same length as the patient graph then the two graphs are 
    multiplied together (element-wise) to get the activated graph.

    Raises:
        ValueError: if the patient graph is not 3D then raise an error.

    Returns:
        np.array: filter array .* patient array.
    """

    if pat_graph.ndim != 3:
        raise ValueError("The input array must be 3-dimensional.")

    vectorized_leaky_relu = np.vectorize(leaky_relu)
    
    filt_num_repeats = pat_graph.shape[0]/max_act_filt.shape[0]


    # To reduce compute repeat the filter to be the same size as the patient graph.
    # That way the two can be directly multiplied rather than sliding.
    filt_repeated = repeat_array_fractional(max_act_filt, filt_num_repeats)

    # Element-wise multiplication for the patient graph and the repeated filter
    filt_times_graph = filt_repeated * pat_graph

    # apply a leaky relu (activation)
    activated_graph = vectorized_leaky_relu(filt_times_graph, 0.01)

    return activated_graph


from setuptools import setup, find_packages


with open("README.md", "r") as f:
    description = f.read()

setup(
    name="tgcnn_act_graph",
    version="0.2",
    packages=find_packages(),
    install_requires=[
        # 
    ],
    long_description=description,
    long_description_content_type="text/markdown",
)


import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
import networkx as nx
import numpy as np

from tgcnn_act_graph import utils, max_act_diff, calculations

def plot_activation_difference(filter_num_col:pd.DataFrame, diff_col:pd.DataFrame):
    """_summary_

    Args:
        filter_num_col (pd.DataFrame): corresponding filter number
        diff_col (pd.DataFrame): difference in activation between positive and negative class
    """
    # Plotting the line chart
    plt.figure(figsize=(10, 6))
    plt.plot(filter_num_col, diff_col, marker='.', linestyle='-')
    plt.xlabel('Filter number')
    plt.ylabel('Difference in activation between positive and negative class')
    plt.xticks(filter_num_col, rotation=45)  # Set x-axis ticks to integer values with rotation
    plt.tight_layout()  # Adjust layout to prevent clipping of labels
    plt.show()


def draw_edge_activated_graph(edges_df:pd.DataFrame, pos_dict:pd.DataFrame):
    """Draw an individual patient graph using NetworkX highlighting which edges might 
    be most associated to the prediction.

    Args:
        edges_df (pd.DataFrame): Dataframe with information about edges including: start and end nodes, whether it's activated,
        weight, and edge_label.
        pos_dict (dictionary): dictionary with x and y coordinates for each node:(x,y)
    """
    # Convert df to list of tuples for Networkx
    edges = []
    for _, row in edges_df.iterrows():
        edge = (row['start_node'], row['end_node'], {'activated': row['activated'], 'weight': row['weight'], 'edge_label': row['time_between']})
        edges.append(edge)

    G = nx.DiGraph()
    G.add_edges_from(edges)


    # Prepare edge colors, widths, and labels based on attributes
    edge_colors = ['red' if G[u][v]['activated'] == 1 else 'grey' for u, v in G.edges()]
    edge_widths = [G[u][v]['weight'] for u, v in G.edges()]
    edge_labels = {(u, v): G[u][v]['edge_label'] for u, v in G.edges()}

    # Draw the graph
    plt.figure(figsize=(12, 8))
    nx.draw(G, pos_dict, with_labels=True, node_size=3000, node_color="lightblue", edge_color=edge_colors, width=edge_widths, font_size=10, font_weight="bold", arrowsize=20)
    nx.draw_networkx_edge_labels(G, pos_dict, edge_labels=edge_labels, font_color='blue', font_size=10, font_weight='bold')

    #plt.title("Graph Visualisation of Patients Pathway and Connections Associated to Hip Replacement")
    plt.show()



def edge_activated_graph(input_tensors:np.array, patient_number:int, filters:np.array, labels:list, verbose:bool=False, show_plot:bool=False):
    """Draw an individual patients graph with red edges where the AI model filters have given high weights, 
    representing which edges the patterns associate to the outcome the most. 

    Args:
        input_tensors (np.array): 4D array of patient graph representations (3D).
        patient_number (int): Patient to represent as a graph.
        filters (np.array): 4D array of filters from the TG-CNN model (3D).
        labels (list): Binary outcome (0 or 1) for patient e.g. no replacement or hip replacement.
        verbose (bool): Print extra statements.
        show_plot (bool): Print the maximum activation per filter plot.
    """

    # 1. Select the patient graph to draw
    patient_graph = utils.select_patient(input_tensors, patient_number)

    # 2. Get the maximum activation for each filter.
    mean_activation_df = max_act_diff.max_act_diff_calc('Hip Replacement', input_tensors, filters, labels, verbose=verbose, show_plot=show_plot)

    # 3. Find the filter with the maximum activation.
        # This should have the strongest effect.
    max_act_filt = utils.get_max_act_filt(mean_activation_df, filters)

    # 4. Element-wise multiplication of the max activation filter by the chosen graph
    act_graph = calculations.get_act_graph_array(patient_graph, max_act_filt)

    # 5. Get the edge features incl. start and end node, activation, weight, time between (edge value)
    edges_df = utils.create_edges_df(patient_graph, act_graph)

    # 6. Create a dataframe with the x coordinates for the nodes
    pos_df = utils.create_position_df(edges_df)

    # 7. Generate a list of possible y locations depending on the number of nodes per visit
    pos_list = utils.generate_pos_sequence(pos_df['max_codes_per_visit'].max())

    # 8. Add to the dataframe to y get coordinates for the nodes based on the number of nodes per timestep
    pos_df = utils.map_y_coord_to_node(pos_df, pos_list)

    # 9. Turn the x and y coordinates into a dictionary that can be read by NetworkX
    pos_dict = utils.create_pos_dict(pos_df)

    # 10. Draw the patient graph with the activated edges
    draw_edge_activated_graph(edges_df, pos_dict)